{
	This unit contains the miscellaneous routines,
	and the body of the program.
}
{ Created in <=2000 - Copyright (C) Joel Yliluoma }

Unit NESMusa1;

Interface

Uses SetJmp;

Const
	PlayBufSize = 2048;

Const
	S3MVersion = $5728; {1320 = ST 3.20}
	TitleStr = 'NESMusa V7.28 (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)'^M^J+
             'Special edition -- compiled just for 8bitbubsy in 2010-03-19 ;)';
  Title: Array[0..Length(TitleStr)]Of Char = TitleStr;

Const
  PatnNotes: Array[0..15, 0..2]Of Char =
   ('C-','C#','D-','D#','E-','F-','F#','G-','G#','A-','A#','B-',
    '??','??','^^','..');
Const
  Name: Array[0..128]Of Char = '';

Const
  BasicGamma = 33;

  AutoTicker: Boolean = False;
  GammaValue: LongInt = 0;
  TakaPerin: Boolean = False;
  AltInstru: Boolean = False;
  AltInstru2:Boolean = False;
  Sustain:   Integer = 0;

Const
  FileLoadDebug: Boolean = False;

Type
  Float = Extended;

Type
  S3MHdr = Record
    Name: Array[0..27]Of Char;
		CtrlZ: Char; {^Z}
    FileType: Byte; {16 = S3M}
    Filler1: Word;
    OrdNum: Word; {Number of orders, is even}
    InsNum: Word; {Number of instruments}
    PatNum: Word; {Number of patterns}
    Flags : Word; {Don't care}
    Version: Word; {ST3.20 = $1320, ST3.01 = $1301}
    SmpForm: Word; {2 = Unsigned samples}
    Ident : Array[0..3]Of Char; {SCRM}
    GVol  : Byte; {Global volume, don't care}
    InSpeed: Byte; {Initial speed}
    InTempo: Byte; {Initial tempo}
    MVol  : Byte; {Master volume, don't care}
    UC    : Byte; {Ultraclick removal, don't care}
    DP    : Byte; {Default channel pan, don't care}
    Filler2: Array[$6..$F]Of Byte;
    Channels:Array[0..31]Of Byte; {Channel settings, 16..31 = adl channels }
  End;

Const
  MAXORD = 127;
  MAXINST = 4;
  MAXPATN = 100;

Var
  Hdr: S3MHdr;
  Orders: Array[0..MAXORD-1]Of Byte;
  OrdVis: Array[0..MAXORD*64-1]Of Boolean;
  PatnPtr: Array[0..MAXPATN-1]Of Word;
  InstPtr: Array[0..MAXPATN-1]Of Word;

Type
  PPatn = ^TPatn;
  TPatnItem = Record
    Len: Word;
    Ptr: PChar;
  End;
  TPatn = Array[0..MAXPATN-1]Of TPatnItem;
Var
  Patn: PPatn;

Procedure Loop;
Procedure Read(s: PChar);
Procedure ArgInit;
Procedure InitLoop;
Procedure Simulate;
Procedure GetSpeed;
Procedure WriteStr(Size: Word; Str: PChar);
Procedure Error(Const s: String);
Procedure Debug(Const s: String);
Function GammaCalculation(Volume: Byte): Byte;

Procedure ListUsage;
Procedure ComprInfo;
Procedure VersInfo;

Procedure Show2; {Plays with beeper}

Var
  Player: Procedure;

Const
	DMA: Byte = 1;
	IRQ: Byte = 5;
  Period: Array[0..11]Of LongInt =
   (907,960,1016,1076, 1140,1208,1280,1356, 1440,1524,1616,1712);
  BasePort: Word = $220;
  GlobalVolume: Byte = 140;

Const
	BufSize: Word = PlayBufSize;
  TicPeriod: Byte = 60;
  Rate = 33075;

Type
  TRivi = Array[0..3]Of Record Note, Volume: Byte End;
  TSong = Array[0..MAXORD*64-1]Of TRivi;
  PSong = ^TSong;
	TOrdRows = Array[0..MAXORD*64-1]Of Word;
	POrdRows = ^TOrdRows;

Var
  Laulu: PSong;
  Riveja: Word;
  Looppaa: Word;
  LastChan: Byte;
	OrdRow: POrdRows;
  DeVo: Array[0..3]Of Byte; {Default volume}
  Spiid: Float;

Var
  QuickExit: JmpBuf;

Procedure NSMP1; {NESMUSA5.OBJ -  112 bytes}
Procedure NSMP2; {NESMUSB5.OBJ -  208 bytes}
Procedure NSMP3; {NESMUSC5.OBJ - 8272 bytes}

Implementation

Uses Objects,WinDos,Strings,Crt,SVGA,DosMem,NesMusa2{$IFNDEF PLAYERONLY},NesMusa4,NesMusa5{$ENDIF};

Procedure NSMP1; External; {$L NESMUSA5.OBJ}
Procedure NSMP2; External; {$L NESMUSB5.OBJ}
Procedure NSMP3; External; {$L NESMUSC5.OBJ}

Procedure Debug(Const s: String); Begin If Debugging Then WriteLn(s)End;

Procedure Error(Const s: String);
Begin
  WriteLn(s);
  LJmp(QuickExit, 1)
End;

Var
  LoopTimer: LongInt;
  LoopMaxi: LongInt;
  LoopVar: Boolean;
Const
  LoopMax: LongInt = 200;
  LoopWarnCount: Byte = 0;

Procedure InitLoop;
Begin
  LoopTimer := MemL[Seg0040:$6C];
  LoopMaxi := LoopMax;
  LoopVar := False
End;

Procedure SendToDos(Size: Word; Str: PChar); Assembler;
Asm
  push ds
   mov ah, 40h
   mov bx, 1
   mov cx, Size
	 lds dx, Str
   int 21h
  pop ds
  jnc @@1
  mov InOutRes, ax
@@1:
  {$IFOPT I+}call InOutCheck{$ENDIF}
End;

Procedure WriteStr(Size: Word; Str: PChar);
Begin
  If Size=0 Then Exit;
  LoopVar := True;

  SendToDos(Size, Str)
End;

Procedure Loop;
Begin
  If LoopVar Then InitLoop
  Else If MemL[Seg0040:$6C]-LoopTimer > LoopMaxi Then
  Begin
    Case LoopWarnCount Of
      0: WriteLn('Buy a faster computer.');
      1: WriteLn('Make a faster computer.');
      2: WriteLn('I hope I''m not boring you.');
      3: WriteLn('Bisqwit likes X-Files, did you know?');
      4: WriteLn('Boo.');
      5: WriteLn('You gave me a big job.');
      6: WriteLn('Ouch.')
    End;
    InitLoop;
    Inc(LoopWarnCount);
    LoopMaxi := MaxLongInt;
    Dec(LoopMax, 50)
  End;
  If KeyPressed Then
  Begin
    If ReadKey=#0 Then ReadKey;
    SendToDos(31, #13'*** Stopped by keyboard user'^M^J);
    Halt
  End
End;

Function GammaCalculation(Volume: Byte): Byte;
	Function SubGamma(v: Byte): Float;
  Begin
    SubGamma :=
      Exp((1 / Sqrt(Exp(GammaValue * Ln(1.135)))) * Ln(v))
  End;
Begin
{ GammaCalculation := Volume;}
  {GammaCalculation := Volume ^ (1 / Sqrt(1.14 ^ GammaValue))}

  If GammaValue=0
    Then GammaCalculation := Volume
    Else If Volume = 0
      Then GammaCalculation := 0
      Else GammaCalculation := Trunc(63 * SubGamma(Volume) / SubGamma(63))

  { a ^ b = exp(b * ln(a)) }
End;

Procedure Read(s: PChar);
Var
  l: LongInt;
  BRead: Word;
  f: TDosStream;
  i, j: Integer;
  Hmm, Hmm2: Byte;
  Tiedosto: Boolean;
  ErrText, ErrWhen: String[80];
  maxrow: Word;
Label (*P1, *)P2, P3, P4;
Begin
  Tiedosto := s[1] = ':';

P2:
  FileMode := 0;

  If Tiedosto Then
  Begin
    If Debugging And FileLoadDebug
      Then Write('Opening s')
      Else Write('S');
    Write('ource file: ');

    f.Init(s, stOpenRead + stDenyWrite);

    If f.Status <> 0 Then
		Begin
      f.Done;
      WriteLn(s);
      Error('*** Trouble opening.')
    End;

    If Debugging Then
    Begin
      BRead := LFN_GetFAttr(s);
      Write('(');
      If (BRead And faReadOnly) <> 0 Then Write('R')Else Write('-');
      If (BRead And faHidden  ) <> 0 Then Write('H')Else Write('-');
      If (BRead And faSysFile ) <> 0 Then Write('S')Else Write('-');
      If (BRead And faArchive ) <> 0 Then Write('A')Else Write('-');
      Write(') ')
    End;

    WriteLn(s)
  End;

  ErrWhen := ' Reading header';
  If FileLoadDebug Then
    If Tiedosto
      Then Debug(ErrWhen)
      Else Debug('Looking for header in memory');

  If Tiedosto Then
  Begin
    f.Read(Hdr, SizeOf(Hdr));
    ErrText := 'Not a S3M file!';
P3: If f.Status = stReadError Then
    Begin
      f.Done;
      Goto P4
    End;
    If f.Status <> stOk Then
    Begin
      Write('*** Error ',f.ErrorInfo);
      Error(ErrWhen)
    End
  End
  Else
    Move(s^, Hdr, SizeOf(Hdr));

  If strncmp(PChar(@Hdr), 'SND'^Z, 4) = 0 Then
	Begin
    If Tiedosto Then f.Done;
    ErrText := 'This is a SND file. Convert it to S3M first.';
    Goto P4
  End;

  Hdr.GVol := 0;
  If(strncmp(Hdr.Ident, 'SCRM', 4) <> 0)Then
  Begin
    If Tiedosto Then f.Done;
    If Not Tiedosto Then
    Begin
      Tiedosto := True;
      Goto P2;
    End;
    ErrText := 'Not a S3M file!';
P4: Error('*** '+ErrText)
  End;
  If Hdr.InsNum > MAXINST Then
		WriteLn('Warning: ','This is possibly not a generated NES music file.');
	If(Hdr.OrdNum > MAXORD )Or(Hdr.PatNum> MAXPATN)Then Write('*** File has too ');
  If(Hdr.OrdNum > MAXORD )Then Begin If Tiedosto Then f.Done; Error('much orders!')End;
  If(Hdr.PatNum > MAXPATN)Then Begin If Tiedosto Then f.Done; Error('much patterns!')End;

  If AutoTicker Then TicPeriod := 24 * Hdr.InTempo Div Hdr.InSpeed Div 60;

  maxrow := 65520 Div SizeOf(TRivi); {Hdr.OrdNum*64;}
  BRead := (maxrow * SizeOf(TRivi) + 15)Shr 4;
  AllocMem('song', Laulu, BRead);
  FillChar(Laulu^, BRead Shl 4, 0);

  If FileLoadDebug Then Debug(' Reading misc things...');

  FillChar(Orders, SizeOf(Orders), 255);

  If Tiedosto Then
  Begin
    f.Seek(SizeOf(Hdr));
    f.Read(Orders, Hdr.OrdNum);
    If f.Status <> stOk Then
    Begin
      ErrWhen := ' Reading orders';
      ErrText := '*** Not enough orders!';
      Goto P3
    End;

    f.Read(InstPtr, Hdr.InsNum*2);
    If f.Status <> stOk Then
    Begin
      ErrWhen := ' Reading instrument pointers';
      ErrText := 'Not enough instrument pointers!';
      Goto P3
    End;

    f.Read(PatnPtr, Hdr.PatNum*2);
    If f.Status <> stOk Then
    Begin
      ErrWhen := ' Reading pattern pointers';
      ErrText := 'Not enough pattern pointers!';
      Goto P3
    End;
  End
  Else
  Begin
    Move(s[SizeOf(Hdr)], Orders, Hdr.OrdNum);
    Move(s[Hdr.InsNum*2+Hdr.OrdNum+SizeOf(Hdr)], PatnPtr, Hdr.PatNum*2)
  End;

  If FileLoadDebug Then Debug(' Reading patterns');

  AllocMem('pattern pointer table', Patn, (SizeOf(TPatnItem)*Hdr.PatNum+15)Shr 4);

  Hmm2 := 0;
  For i := 0 To Hdr.PatNum-1 Do If PatnPtr[i] = 0 Then Inc(Hmm2);
  If Hmm2 <> 0 Then
  Begin
    If FileLoadDebug And Debugging Then
    Begin
      Write('Pattern');
      If Hmm2 <> 1 Then Write('s');
      Write(' ')
    End;
    Hmm := 0;
    For i := 0 To Hdr.PatNum-1 Do
      If PatnPtr[i] = 0 Then
      Begin
        Patn^[i].Len := 0;
        Patn^[i].Ptr := Nil;
        If FileLoadDebug And Debugging Then
        Begin
					Inc(Hmm);
          If Hmm = Hmm2 Then Write(i)
          Else If(Hmm = Hmm2-1)And(Hmm2 <> 1)Then Write(i,' and ')
          Else Write(i, ', ')
        End
      End;
    If FileLoadDebug And Debugging Then
    Begin
      If Hmm2 = 1 Then Write(' is')Else Write(' are');
      Write(' empty! What the ');
      If Hmm2 > 1 Then Write('gigantic ')
    End;
    If FileLoadDebug Then Debug('waste of timespace!')
  End;

  For i := 0 To Hdr.InsNum-1 Do
    If i<=High(Devo)Then
    Begin
			DeVo[i] := 0; {Default volumes...}
      If InstPtr[i] <> 0 Then
        If Tiedosto Then
        Begin
          f.Seek(InstPtr[i]*16 + $1C);
          f.Read(DeVo[i], 1);
          If f.Status <> stOk Then
          Begin
            ErrWhen := ' Reading instrument default volumes';
            ErrText := 'Not enough instruments(?)';
            Goto P3
          End
        End
        Else
          Move(s[InstPtr[i]*16 + $1C], DeVo[i], 1)
    End;

  l := 0;
  For i := 0 To Hdr.PatNum-1 Do
  Begin
    Loop;
    If PatnPtr[i] <> 0 Then
    Begin
      If Tiedosto Then
      Begin
        f.Seek(PatnPtr[i] * 16);
        f.Read(Patn^[i].Len, 2);
				If f.Status <> stOk Then
        Begin
          ErrWhen := ' Reading patterns';
          ErrText := 'Not enough patterns!';
          Goto P3
        End
      End
      Else
        Move(s[PatnPtr[i]*16], Patn^[i].Len, 2);

      Dec(Patn^[i].Len, 2);

      If Patn^[i].Len > 1344 Then
        If Debugging And FileLoadDebug Then
          WriteLn('Suspicious pattern length: ', Patn^[i].Len,' in pattern ',i);

      AllocMem('', Patn^[i].Ptr, (Patn^[i].Len+15)Shr 4);
      Inc(l, Patn^[i].Len);

      If Tiedosto Then
      Begin
        f.Read(Patn^[i].Ptr^, Patn^[i].Len);
        If f.Status <> stOk Then
        Begin
          f.Done;

          Write(
            '*** Not enough pattern data in pattern ',i,'!'^M^J+
            '*** Wanted ', Patn^[i].Len);
          Error(' bytes, read less.')
        End
      End
      Else
        Move(s[PatnPtr[i]*16+2], Patn^[i].Ptr^, Patn^[i].Len)
    End
  End;

  If FileLoadDebug Then Debug('Closing source file');
  If Debugging And DosMemDebug Then
    WriteLn('At least ',l,' bytes allocated for patterns');

  If Tiedosto Then f.Done
End;

Procedure Simulate;
Var
	a: Byte;
	i: Integer;
	Ptr: PChar;
	Rivi: TRivi;
	Idle: Boolean;
	Pat, Row, Order: Byte;
	Skip, JumpRow, JumpOrd: Byte;
	LisaaRiveja, Chan, Note, Volume, Command, Info: Byte;
	maxrow: word;
	LoopBegin, w: Word;
	LoopTimes: Byte;
Label P1, P2, OutOfRows;
Begin
	Debug('Inquiring into the song...');
	maxrow := 65520 Div SizeOf(TRivi); {Hdr.OrdNum*64;}
	AllocMem('ordrow table', OrdRow, (15+maxrow*2)Shr 4);

	FillChar(OrdVis, SizeOf(OrdVis), 0);

	Idle := False;
	LastChan := 0;
	Riveja := 0;
	Order := 0;
	Skip := 0;
	Repeat
P1: Loop;
		While Orders[Order] = 254 Do Inc(Order);
		If Orders[Order] = 255 Then
		Begin
			If Debugging Then WriteLn(' Song ends at order #',Order);
			Break
		End;
		Pat := Orders[Order]Mod MAXPATN;
		Ptr := Patn^[Pat].Ptr;

		If FileLoadDebug Then
			If Ptr = Nil Then
				If Idle Then
					WriteLn(' Adding more idle (pattern ',Pat,').')
				Else
				Begin
					WriteLn(' Adding 64 rows of idle (pattern ',Pat,')... Nice.');
					Idle := True
				End;

		LoopBegin := Riveja;
		LoopTimes := 0;

		For Row := 0 To 63 Do
		Begin
			If OrdVis[Order*64+Row] Then
			Begin
				For i := 0 To Riveja-1 Do
					If Order Shl 8 + Row = OrdRow^[i]Then
					Begin
						Looppaa := i;
						If FileLoadDebug Then
              WriteLn('Song repeat point is at row ',Looppaa);
            Goto P2
          End;
        Error('*** Internal error: Repeat point not found.');
        Looppaa := 0;
        Goto P2
      End;

      If Skip=0 Then OrdVis[Order*64+Row] := True;

      If Riveja > 0
        Then Rivi := Laulu^[Riveja-1]
				Else FillChar(Rivi, SizeOf(Rivi), 0);

      JumpOrd := 255;
      JumpRow := 255;
      LisaaRiveja := 0;
      Repeat
        Loop;
        If Ptr = Nil Then Break;
        a := Ord(Ptr^); Inc(Ptr);
        If a = 0 Then Break;
        Chan := a And 31;

        If Chan >= 4 Then
        Begin
          If Debugging Then
            Write(
              'Found stuff on channel ',Chan,
              ' in pattern ',Pat,'. ');
					If(Hdr.Channels[Chan]=255)Then
          Begin
						Debug(
							#13^J'The channel does not exist. Seems like a bug.'^M^J+
							'Maybe the song did end a long time ago... At least, it is over NOW.');
						Goto P2
          End;
				End;

        Note := 255;
        Volume := 255;
        Command := 255;
				If((a And 32) <> 0)Then
				Begin
					Note := Ord(Ptr^); Inc(Ptr);
					If Ptr^ <> #0 Then Volume := GammaCalculation(DeVo[Ord(Ptr^)-1]);
          Inc(Ptr)
        End;
        If((a And 64) <> 0)Then
        Begin
          Volume := GammaCalculation(Ord(Ptr^));
          Inc(Ptr)
        End;
        If((a And 128) <> 0)Then
        Begin
          Command := Ord(Ptr^)Or 64; Inc(Ptr);
          Info    := Ord(Ptr^);      Inc(Ptr)
				End;

        If Note = 254 Then
        Begin
          Volume := 0;
          Note := 255
        End;

        If Note <> 255 Then
        Begin
          If LastChan < Chan Then LastChan := Chan;
          Rivi[Chan].Note  := Note
        End;
        If Volume<>255 Then Rivi[Chan].Volume:= Volume;

        Case Chr(Command) Of
          'S':
						Case Info Shr 4 Of
							$E: LisaaRiveja := Info-$E0;
							$B:
								If Info=$B0 Then
									LoopBegin := Riveja
								Else
									LoopTimes := Info-$B0;
						End;
					'C': JumpRow := (Info Shr 4)*10 + (Info And 15);
					'B':
						JumpOrd := Info;
				End
			Until False;

			If Skip <> 0 Then
				Dec(Skip)
			Else
			Begin
				For a := 0 To LisaaRiveja Do
				Begin
					OrdRow^[Riveja] := Order Shl 8 + Row;
					Laulu^[Riveja] := Rivi;
					If Riveja >= maxrow-2 Then Goto OutOfRows;
					Inc(Riveja)
				End;

				If LoopTimes <> 0 Then
				Begin
					LoopBegin := Riveja - LoopBegin; { # of rows to loop }
					For w := 1 To LongInt(LoopBegin) * LoopTimes Do
					Begin
						OrdRow^[Riveja] := Order Shl 8 + Row;
						Laulu^[Riveja] := Laulu^[Riveja-LoopBegin];
						If Riveja >= maxrow-2 Then Goto OutOfRows;
						Inc(Riveja)
					End;
					LoopTimes := 0;
					LoopBegin := Riveja
				End;

				If JumpOrd <> 255 Then
				Begin
					Order := JumpOrd;
					If JumpRow <> 255 Then Skip := JumpRow;
					Goto P1
				End;
				If JumpRow <> 255 Then
				Begin
					Inc(Order);
					Skip := JumpRow;
					Goto P1
				End
			End
		End;
OutOfRows:
		Inc(Order)
	Until False;
P2:
	If Debugging Then
	Begin
		Write(' Song consists of ',Riveja,' rows');
		If Riveja >= maxrow-2 Then Write(' - CUT.');
    WriteLn
  End;

	If Debugging And DosMemDebug Then
		WriteLn(' Song length is ',Riveja*SizeOf(TRivi),' bytes');

  ReSize(Laulu, (Riveja*SizeOf(TRivi)+15)Shr 4);

  UnAllocMem(OrdRow);
  UnAllocMem(Patn);

  If DosMemDebug Then Debug('Disposing patterns');
  For i := 0 To Integer(Hdr.PatNum)-1 Do
    With Patn^[i]Do
      If Ptr <> Nil Then
      Begin
        UnAllocMem(Ptr);
        Ptr := Nil;
        Len := 0
      End
End;

Procedure GetSpeed;
Const
  Incr = 0.01;
  Min  = 3.00;
  Max  =15.00;
Var
  i, j: Float;
  Test: Float;
  Chan, Vol: Byte;
	Len: Word;
  Match: Word;
  Best3, Best, Best2, BSpiid: Float;
  Div3, Divi, Div2: Integer;

  Hmm, Hmm2: Byte;
Begin
  Debug('Finding the song tempo...');

  Best2 := 0;
  Chan := 2;
  Repeat
    Test := Min;
    Best := 0;
    If Debugging Then Write('Score ',Best:0:4,#13);
    Repeat
      { Hypitn jonnekin kohden svelm }
      i := Riveja Div 4;
      Vol := 63;
      While Laulu^[Round(i)][Chan].Volume <= Vol Do
      Begin
        Loop;
        Vol := Laulu^[Round(i)][Chan].Volume;
        i := i + 1
      End;
      Len := 0;
      Match := 0;

      If i >= Riveja Then
        Len := 65535
      Else
      Begin
        j := i;
        While i < Riveja-1 Do
        Begin
          Loop;
          Inc(Len);
          If Laulu^[Round(i)][Chan].Volume
          > Laulu^[Round(i-1)][Chan].Volume Then
            Inc(Match);
          i := i + Test;
          Inc(Len)
        End
      End;
      If Match/Len > Best Then
			Begin
        Best := Match/Len;
        If Debugging Then Write('Score ',Best:0:4,#13);
        Spiid := Test
      End;
      Test := Test + Incr;
      Loop
    Until Test > Max+Incr;

    If Debugging Then
    Begin
      If Chan = 2 Then
        Write(
          'Tested tempos between ',
          Min:0:2,'..',Max:0:2,
          '; song tempo is ',Spiid:0:2,
          ' (score ',Best:0:4,')')
      Else
        Write(
          'Tempo ',Spiid:0:2,
          ', score ',Best:0:4);
    End;

    If Best >= 0.22 Then
    Begin
      If Chan <> 2
        Then Debug('; that''s enough.')
        Else Debug('');
      Break
    End;

    If Debugging Then
      If Chan <> 2 Then Write(' (According the data on channel ',Chan,')');
    Debug('.');

    If Best > Best2 Then
    Begin
      Best2 := Best;
      BSpiid := Spiid
    End;
    If Chan=2 Then
    Begin
      Debug('Poor score. Testing another channel.');
      Chan := 3
    End
		Else If Chan=3 Then
    Begin
      Debug('The score was still quite low. Testing one channel more.');
      Chan := 1
    End
    Else If Chan=1 Then
    Begin
      Debug('Score is still low. Testing the last channel.');
      Chan := 0
    End
    Else
    Begin
      Debug('Maybe the song is random, but I give up now.');
      Spiid := BSpiid;
      If Debugging Then
        WriteLn('Tempo ',Spiid:0:2,' had the biggest score (',Best2:0:4,').');
      Break
    End
  Until False;

  Best3 := 1;

  Spiid := Round(Spiid * 100) / 100;

  If Spiid = 10.67 Then Spiid := Spiid/6;
  If Spiid = 10.00 Then Spiid := 2;

  If(Frac(Spiid) < 0.4)Then
  Begin
    Hmm2 := 0;
    For Divi := 2 To Round(Spiid)Div 2 Do
      If Round(Frac(Spiid / Divi)*100)/100 < 0.35 Then Inc(Hmm2);
    If Hmm2 <> 0 Then
    Begin
      If Debugging Then
      Begin
        Write('Tempo may be divided ');
        If Hmm2 = 1 Then Write('only ');
        Write('by ')
      End;
      Hmm := 0;
      Div2 := 0;
      For Divi := 2 To Round(Spiid)Div 2 Do
      Begin
        Loop;
				If Round(Frac(Spiid / Divi)*100)/100 < 0.35 Then
        Begin
          If(Round(Frac(Spiid / Divi)*100)/100 < Best3)
          Or((Round(Frac(Spiid / Divi)*100)/100 < 0.01)
           And(Divi < Round(Spiid)Div 2))Then
          Begin
            Div3 := Divi;
            Best3 := Round(Frac(Spiid / Divi)*100)/100
          End;
          Inc(Hmm);
          If Spiid/Divi > 2.2 Then Div2 := Divi;
          If Debugging Then
            If Hmm = Hmm2 Then Write(Divi)
            Else If(Hmm = Hmm2-1)And(Hmm2 <> 1)Then Write(Divi,' and ')
            Else Write(Divi, ', ')
        End
      End;
      If Div2 = 0 Then
        If Debugging Then Write(', doing nothing')Else
      Else
      Begin
        If Best3 < 0.01 Then Div2 := Div3;
        If Debugging Then
          If Hmm2 = 1
            Then Write('; selection is obvious..')
            Else Write(', selecting ',Div2);
        Spiid := Spiid / Div2;
      End;
      Debug('.')
    End
  End
  Else If Debugging Then Exit;
  WriteLn('Song tempo is ',Spiid:0:2)
End;

Procedure Show2;
Var
  j: Byte;
  Herz: LongInt;
  Note: Word;
  i: Word;
Const
  Allowed: Set Of Byte = [0,1,2];
  Rep = 1;
Var
	Time: Array[0..4]Of Float;
  Vol, Chans, k: Integer;
Begin
  If TakaPerin
    Then i := Riveja-1
    Else i := 0;
  Repeat
    Chans := 0;
    For k := 0 To 2 Do If k In Allowed Then Inc(Chans);
    For k := 1 To Rep Do
    Begin
      Write(i:4,':  ');

      Time[3] := 1000/54/Rep;
      Time[4] := Time[3];
      For j := 0 To 2 Do
      Begin
        Note := Laulu^[i][j].Note And 15;
        Write(PatnNotes[Note]);

        If Laulu^[i][j].Note < 254
          Then Write(Laulu^[i][j].Note Shr 4)
          Else Write('.');

        Write(Laulu^[i][j].Volume:3, ' | ');

        If(Laulu^[i][j].Volume=0)And(j In Allowed)Then NoSound
        Else If(Laulu^[i][j].Note < 254)And(j In Allowed)Then
        Begin
          Herz := Period[Note] * (1 Shl (Laulu^[i][j].Note Shr 4));
          Sound(Herz Div 20)
        End;
        If j In Allowed Then
        Begin
          Vol := Laulu^[i][j].Volume;
          If Vol>47 Then Vol := 47;
          Time[j] := Time[3] * Vol / (47*Chans);
          Time[4] := Time[4] - Time[j];
          Delay(Round(Time[j]))
        End
      End;
      Write(#13);
      NoSound;
      Delay(Round(Time[4]))
    End;

    Write(#10);

    If KeyPressed Then
      Case ReadKey Of
        '1': If 0 In Allowed
            Then Allowed := Allowed - [0]
            Else Allowed := Allowed + [0];
        '2': If 1 In Allowed
            Then Allowed := Allowed - [1]
            Else Allowed := Allowed + [1];
        '3': If 2 In Allowed
            Then Allowed := Allowed - [2]
            Else Allowed := Allowed + [2];
        Else
          Break
      End;
    If TakaPerin Then
    Begin
      Dec(i{, Round(Spiid)});
      If i < Looppaa Then i := Riveja-1
    End
    Else
    Begin
      Inc(i{, Round(Spiid)});
      If i >= Riveja Then i := Looppaa
    End
  Until False;
End;

Procedure EnvInit(AIParam: Boolean);
Var
  i: Word;
  a, p: PChar;
  Code: Integer;
  Buf: Array[0..128]Of Char;
Begin
  p := GetEnvVar('BLASTER');
  If p<>Nil Then
  Begin
    If AIParam Then
      Debug('BLASTER environment variable found, but skipping it')
    Else
    Begin
      Debug('Found BLASTER environment setting!');

      a := StrScan(p, 'D');
			If a=Nil Then Debug('*** Invalid BLASTER setting; DMA (Dx) not found.');
			Inc(a);
			If(a[0]>'3')Or(a[0]<'0')Or((a[1]>='0')And(a[1]<='9'))Then
				Error('*** Invalid BLASTER setting: Only DMAs 0..3 supported.');
			DMA := Ord(a[0]) - Ord('0');

			a := StrScan(p, 'I');
			If a=Nil Then Error('*** Invalid BLASTER setting; IRQ (Ix) not found.');
			Inc(a);
			If(a[0]>'7')Or(a[0]<'0')Or((a[1]>='0')And(a[1]<='9'))Then
				Error('*** Invalid BLASTER setting: Only IRQs 0..7 supported.');
			IRQ := Ord(a[0]) - Ord('0');

      p := StrScan(p, 'A');
      If p=Nil Then Error('*** Invalid BLASTER setting. Port (Axxx) not found.');
      Inc(p);

      strcpy(Buf, p);
      a := StrScan(Buf, ' ');
      If a<>Nil Then Buf[a-Buf] := #0;
      Val(Buf, BasePort, Code);
      Case BasePort Of
				200,210,220,230,240,250,260:
          BasePort := $200 + ((BasePort Div 10)Mod 10)Shl 4;
        Else
          Error('*** Invalid BLASTER port setting.');
      End
    End
  End
End;

Procedure ListUsage;
Begin
  WriteLn(
'                           NESMUSA OPTIONS'^J);

  WriteLn(
'                              -PLAYERS-'^M^J^J,

' -P','  Play music (','Sound Blaster',') and execute shell   '^M^J,
' -W','  Play music (','Sound Blaster',') and show info (bare)'^M^J,
' -B','  Play music (','PC beeper) - not good quality.     '^J);

{$IFNDEF PLAYERONLY}
	WriteLn(
'                             -GENERATORS-'^M^J^J,

' -G  Generates Turbo Pascal',' sources for a simple player.'^M^J,
' Program plays with ':24,'Sound Blaster',' using DMA. Data ','doublecompressed.'^M^J,
' -Q# Generates QuickBasic',' sources for a simple player.'^M^J,
'   P','  Program plays with "PLAY" statement - not good quality.'^M^J+
'   F','  Program plays',' using FM synthesis. ','Sounds acceptable. Data ','compressed.'^M^J,
'   M','  Multiple songs in one file',' using FM synthesis. ',#13^J+
'      Sounds acceptable. Data ','doublecompressed.'^M^J,
'   S','  Program plays with ','Sound Blaster',' using timer,'^M^J+
'      requires bisqwit.bi and bisqwit.qlb or bisqwit.lib. Data ','compressed.'^M^J,
' -A  Generates assembler',' sources for a simple player.'^M^J,
' Program plays with ':24,'Sound Blaster',' using DMA. Data ','compressed.'^M^J,
' -J1',' Generates a compressed',' song file in OBJ format.'^M^J,
' -J2',' Generates a doublecompressed',' song file in OBJ format.'^M^J,
' -J3',' Generates a doublecompressed',' song library in OBJ format.'^M^J);
{$ENDIF}

	WriteLn(
'                           -FILE OPERATIONS-'^M^J^J,

' -?  This help text to stdout.'^M^J,
' -H# Other help pages'^M^J+
'   D Display decompression method documents to stdout.'^M^J,
'   V Version history.'^M^J
{$IFNDEF PLAYERONLY},
' -E','  Convert a ','SN','D file to a S3M file.'^M^J,
' -E','  Convert a ','MO','D file to a S3M file.'^M^J,
' -E  Rewrite the S3M file. Optimizes the'^M^J+
'     file for size and updates instruments.'^J
{$ENDIF}
);

	WriteLn(
'                         -HARDWARE CONFIGURATION-'^M^J^J,

' -2','  Selects SB port 2','20h (default)      -3','  Selects IRQ ','2 (notice -3, not -2)'^M^J,
' -4','  Selects SB port 2','40h                -5','  Selects IRQ ','5 (default)'^M^J,
' DMA channel 1 always used.              -7','  Selects IRQ ','7 '^J);

{$IFNDEF PLAYERONLY}
	WriteLn(
'                            -OPTIMIZATION-'^M^J^J,

' -1  Disable optimizing (-QP,-QF,-QS,-QM,-I,-E,-QP,-J2,-J3)'^M^J+
' -D  Increase detail skipping level. (-E). (Even one -D skips a LOT.)'^M^J,
' -L  Brainless working mode (-A,-E,-G,-J,-QF,-QP,-QM,-QS,-J1,-J2,-J3)'^M^J,
'     (Not recommended!)'^M^J+
' -KL Disable SBx optimization (-E) (it should be working, however)'^M^J,
' -KN Use full patterns (disallows Cxx command)'^M^J+
' -KD Skip volume details (-E) (see also -D)'^M^J+
' -Q# Skip # ticks at beginning of .SND file (-E) (default: 0)'^M^J+
' -O  Force additional optimizing (-E,-QP,-J2,-J3)'^J);
{$ENDIF}

	WriteLn(
'                                -OTHER-'^M^J^J,

' -U  Add sustain (-E). Usually not usable.'^M^J+
' -N  Maximize volume (','current value: ',GlobalVolume/2.55:0:0,'%) (-P,-W,-A)'^M^J,
' -R# Gamma/contrast level (0=normal, 100=maximum contrast, ','current value: ',GammaValue,'%)'^M^J,
'     (If -R not specified, -QM,-QF options cause automatic value 35)'^M^J,
' -S1 Selects alternative melodic instrument set.'^M^J,
'     (Lowers C2SPD and lowers instru 1 volume)'^M^J,
' -S2 Selects alternative noise instrument set. (More volume)'^M^J+
' -F  Increases octave by 1 (-P,-W,-G,-Q,-A,-C)'^M^J+
' -T  Plays backwards (-P,-B,-A,-W)'^M^J,
' -V  Increases verbosity level by 1.'^M^J+
'     Level 1 is recommended, 0 is the default.'^M^J+
' -K# Change tick speed to 40 + 4 * #. Default = ', (TicPeriod-40)Shr 2, ' (',TicPeriod,' ticks/sec).'^M^J+
'   A Automatic tick speed (24*Txx/Axx/60) (-P,-W,-G,-QF,-QS,-QM,-A,-E)'^J);

	WriteLn(
'                                -NOTES'^M^J^J,

' Sustain mode is usually not recommended.'^M^J+
' Long parameter lines can be moved to a @file.'^M^J+
' Program has been tested yet only with A220 I5.'^M^J+
' Options are case insensitive... Most, I think.'^M^J,
' Long filenames are now partially supported where the operating system allows.'^M^J+
' Switches -A,-E,-G,-J1,-J2,-J3,-P,-W,-B,-QF,-QM,-QS,-QP are each-others-off-closing.'^M^J+
' Should you have any comments, please them to bisqwit@iki.fi.'^J);

	WriteLn(
		#10+
		' Although this program is large, I am sorry it still cannot'^M^J+
		' read the music directly from the .NES -files. -- Bisqwit')
End;

Procedure ComprInfo;
Begin
	WriteLn(
'                           -DECOMPRESSION METHODS-'^M^J^J,

' DECOMPRESS NORMALLY COMPRESSED DATA:'^M^J+
'   Read word -> Row count           ',' ## May nonexist if optimized'^M^J,
'   Read word -> Loop jump row number',' ## May nonexist if optimized'^M^J,
'   Chan <- 0'^M^J+
'   Start loop'^M^J+
'     ThisRow <- 0'^M^J+
'     Start loop'^M^J,
'       Read byte -> b'^M^J,
'       If b <= 128 Then'^M^J+
'         Song[ThisRow][Chan].Note <- b           ## Halftone number '^M^J+
'         Read byte -> Song[ThisRow][Chan].Volume ## Ranges 0..15'^M^J+
'         Add 1 to ThisRow'^M^J+
'       Else'^M^J,
'         While b>128 Do Loop this:'^M^J+
'           Song[ThisRow][Chan] <- Song[ThisRow-1][Chan]'^M^J,
'           Add 1 to ThisRow, Subtract 1 from b'^M^J,
'         Loop end'^M^J+
'     Loop end when w >= Row count'^M^J+
'     Add 1 to Chan'^M^J+
'   Loop end when Chan == Channel count ## Usually 4, optimization may modify',
#13^J^J,

' DECOMPRESS DOUBLECOMPRESSED DATA:'^M^J+
'   Read word -> Row count           ',' ## May nonexist if optimized'^M^J,
'   Read word -> Loop jump row number',' ## May nonexist if optimized'^M^J,
'   Read byte -> Tick speed (Hz)     ',' ## May nonexist if optimized'^M^J,
'   Chan <- 0'^M^J+
'   Start loop'^M^J+
'     ThisRow <- 0'^M^J+
'     Start loop'^M^J,
'       Read byte -> ','j  ## j and b are 16 bit integers.'^M^J+
'       Read byte -> ','b  ## zeroes the high byte of word.'^M^J,
'       If j == 255 Then'^M^J,
'         j <- b'^M^J+
'         Read byte -> b'^M^J,
'         b <- b + (j And $C0)Shl 2 ## Move two top bits of j to b'^M^J,
'         j <- j And 63             ## So that j will become 6-bit'^M^J,
'         While j>0 Do Loop this:'^M^J+
'           Song[ThisRow][Chan] <- Song[ThisRow-b][Chan]'^M^J+
'           Subtract 1 from j, Add 1 to ThisRow'^M^J+
'         Loop end'^M^J+
'       Else'^M^J,
'         Song[ThisRow][Chan].','Note   <- j'^M^J,
'         Song[ThisRow][Chan].','Volume <- b ## Ranges 0..63'^M^J,
'         Add 1 to ThisRow'^M^J,
'     Loop end when w >= Row count'^M^J+
'     Add 1 to Chan'^M^J+
'   Loop end when Chan == Channel count ## Usually 4, optimization may modify');
End;

Procedure VersInfo;
Begin
	WriteLn(
'                      -VERSION HISTORY, NEWEST FIRST-'^M^J,
#10+
' v7.27:'^M^J+
'   19.12.99 Fixed sample amplitudes and cutted nsmp3 to 4096 bytes'^M^J+
'   01.12.99 Started porting to C or C++.'^M^J,
' v7.26:'^M^J+
'   10.12.99 Finally fixed the SBx optimization'^M^J,
' v7.23:'^M^J+
'   30.12.98 Added -q# switch (skip snd rows)'^M^J+
'   30.12.98 Added -kn switch (disable Cxx)'^M^J+
'   30.12.98 Fixed documentation error in -hd'^M^J,
' v7.22:'^M^J+
'   19.07.98 Added -kd switch (volume detail skipping)..'^M^J,
'   19.07.98 Added SBx optimization (disablable with -kl switch).'^M^J,
'   19.07.98 Added -d switch (detail level decreasing).'^M^J,
'   18.07.98 Fixed crt delayloop init.'^M^J,
' v7.21:'^M^J+
'   12.06.98 Breaking with keyboard should now be working everywhere.'^M^J,
'   12.06.98 Finished the(gamma/contrast level switch (-R).'^M^J,
'            Now FM generations get level 35 automatically.'^M^J,
'   12.06.98 Removed -h switch, renamed -dd to -hd'^M^J,
'   12.06.98 Added -hv switch (version history)'^M^J,
'   12.06.98 Fixed "sound does not have repeat point" error message'^M^J,
#10+
' v7.20:'^M^J+
'   11.06.98 Improved the doublecompression'^M^J+
'   11.06.98 Added -dd switch (compression methods documentation)'^M^J+
'   11.06.98 Added @listfilehavinglongname support on parameter line'^M^J+
'   10.06.98 Added autoticker and tick speed change (-K#) option'^M^J+
#10+
' v7.15:'^M^J+
'            First release on http://www.nettipaja.fi/Bisqwit/files/');
End;

Procedure ArgInit;
Var
	Buf: Array[0..128]Of Char;
  AIParam, Warned,
  WarnedR1, WarnedR2, WarnedR3, UsedR,
  WarnedO, WarnedL,
  WarnedQ, WarnedJ,
  WarnedS, WarnedK: Boolean;
  a, i: Integer;
  p: PChar;
Begin
	AIParam := False;
  Warned  := False;
  @Player := Nil;
	WarnedJ := False;
  WarnedL := False;
  WarnedK := False;
	WarnedO := False;
	WarnedQ := False;
  WarnedS := False;
  WarnedR1 := False;
  WarnedR2 := False;
  WarnedR3 := False;
  UsedR := True;

  DosMemDebug := False;

  InitArgLoop;

  Repeat
    If GetNextArg(Buf, High(Buf)) = Nil Then Break;

    p := Buf;
    If(Buf[0]='-')Or(Buf[0]='/')Then
      While p^ <> #0 Do
      Begin
        Case UpCase(p^) Of
					'-','/': ;
					{$IFNDEF PLAYERONLY}
					'1': Optimizing := False;
					{$ENDIF}
          '2': Begin AIParam := True; BasePort := $220 End;
          '4': Begin AIParam := True; BasePort := $240 End;
					'3': Begin AIParam := True; IRQ := 3         End;
					'5': Begin AIParam := True; IRQ := 5         End;
					'7': Begin AIParam := True; IRQ := 7         End;
					{$IFNDEF PLAYERONLY}
					'A': Player := Show8;
					{$ENDIF}
					'B': Player := Show2;
					{$IFNDEF PLAYERONLY}
					'D': Inc(NoDetails);
					'E': Player := Converter;
					{$ENDIF}
					'F': For a := 0 To 11 Do Inc(Period[a], Period[a]);
					{$IFNDEF PLAYERONLY}
					'G': Player := Show4;
					{$ENDIF}

					'H':
					Begin
						Inc(p);
								 If UpCase(p^) = 'D' Then Player := ComprInfo
						Else If UpCase(p^) = 'V' Then Player := VersInfo
						Else Player := ListUsage
					End;

					'?': Player := ListUsage;
					{$IFNDEF PLAYERONLY}
					'J':
					Begin
						Inc(p);
								 If UpCase(p^) = '1' Then Player := Show7
						Else If UpCase(p^) = '2' Then Player := Show9
						Else If UpCase(p^) = '3' Then Player := Show10
						Else
						Begin
							If Not WarnedJ Then
								WriteLn('Warning: ','Missing/invalid parameter for/of -','J','. Skipping.');
							WarnedJ := True;
							Dec(p)
						End
					End;
					{$ENDIF}
					'K':
					Begin
						Inc(p);
						If UpCase(p^) = 'A' Then AutoTicker := True
						{$IFNDEF PLAYERONLY}
						Else If UpCase(p^) = 'D' Then SkipVolDet := True
						Else If UpCase(p^) = 'L' Then EnableSBx := False
						Else If UpCase(p^) = 'N' Then EnableCxx := False
						{$ENDIF}
						Else If(p^ < '0')Or(p^ > '9')Then
						Begin
							If Not WarnedK Then
								WriteLn('Warning: ','Missing/invalid parameter for/of -','K','. Skipping.');

							WarnedK := True;
							Dec(p)
						End
						Else
							TicPeriod := 40 + 4 * (Ord(p^) - Ord('0'))
					End;
					{$IFNDEF PLAYERONLY}
					'L':
					Begin
						If NoBrains Then
						Begin
							If Not WarnedL Then
								WriteLn('Warning: ','Only brainless people use more than one -L','. Skipping extra.');
						End
						Else
						Begin
							WriteLn('Warning: ','Entering NoBrains working mode, watch out!');
							NoBrains := True
						End
					End;
					{$ENDIF}
					'N': GlobalVolume := 255;
					{$IFNDEF PLAYERONLY}
					'O':
					Begin
						If MustTryMore Then
						Begin
							If Not WarnedO Then
								WriteLn('Warning: ','You need only one -O to set maximum optimization','. Skipping extra.');
							WarnedO := True
						End;

						MustTryMore := True
					End;
					{$ENDIF}
					'P': Player := Show3;
					{$IFNDEF PLAYERONLY}
					'Q':
					Begin
						Inc(p);
								 If UpCase(p^) = 'P' Then Player := Show6
						Else If UpCase(p^) = 'F' Then Player := Show11
						Else If UpCase(p^) = 'M' Then Player := Show14
						Else If UpCase(p^) = 'S' Then Player := Show13
						Else If(p^ >= '0')And(p^ <= '9')Then
						Begin
							Repeat
								SNDSkip := SNDSkip * 10 + (Ord(p^)-Ord('0'));
								Inc(p)
							Until(p^ < '0')Or(p^ > '9');
							Dec(p)
						End
						Else
						Begin
							If Not WarnedQ Then
								WriteLn('Warning: ','Missing/invalid parameter for/of -','Q','. Skipping.');
							WarnedQ := True;
							Dec(p)
						End
					End;
					{$ENDIF}
					'R':
          Begin
            Inc(p);
            GammaValue := 0;
            If(p^ < '0')Or(p^ > '9')Then
            Begin
              If Not WarnedR1 Then
                WriteLn('Warning: ','Missing/invalid parameter for/of -','R','. Using default value, ', GammaValue);

              WarnedR1 := True;
              Dec(p)
            End
            Else
            Begin
              Repeat
                GammaValue := GammaValue * 10 + (Ord(p^)-Ord('0'));
                Inc(p)
              Until(p^ < '0')Or(p^ > '9');
              If p^ = '%' Then
              Begin
                If Not WarnedR2 Then
                  WriteLn('Warning: ','% mark not needed after -R.');
                WarnedR2 := True
              End
              Else
                Dec(p);

              If(GammaValue < 0)Or(GammaValue > 100)Then
              Begin
                GammaValue := 0;
                If Not WarnedR3 Then
                  WriteLn('Warning: ','-R parameter not in allowed range','. Using default value, ', GammaValue);
                WarnedR3 := True;
              End
              Else
              Begin
                UsedR := False;
                WarnedR3 := False;
								WarnedR1 := False
							End
            End
          End;
          'S':
          Begin
            Inc(p);
                 If UpCase(p^) = '1' Then AltInstru := True
            Else If UpCase(p^) = '2' Then AltInstru2 := True
            Else
            Begin
              If Not WarnedS Then
                WriteLn('Warning: ','Missing/invalid parameter for/of -','S','. Skipping.');
              WarnedS := True;
              Dec(p)
            End
          End;
          'T': TakaPerin := True;
          'U': Inc(Sustain);
          'V':
            If Verbose
              Then If FileLoadDebug
                Then DosMemDebug := True
                Else FileLoadDebug := True
              Else Verbose := True;

          'W': Player := Show5;

          Else If Not Warned Then
          Begin
            WriteLn('Warning: ','Skipping unknown options.');
            Warned := True
          End;
        End;
        Inc(p)
      End
  Until False;

  If @Player = Nil Then
    Error('*** You need to specify one of the options below:'^M^J+
          '    -A,-E,-G,-J1,-J2,-J3,-P,-W,-B,-QF,-QM,-QS,-QP');

	{$IFNDEF PLAYERONLY}
	If MustTryMore And Not Optimizing Then
		Error('*** You are trying to make me skitsofrenic, huh? (-O and -1 options together)');

	If @Player = @Converter Then
	Begin
		If Not AutoTicker Then
			WriteLn('Warning: ','It is',' recommended to use ','AutoTicker (-KA',') option with -','E','. Not set.');
		If Optimizing And Not EnableSBx Then
			WriteLn('Warning: ','It is not',' recommended to use ','disable-SBx (-KL',') option with -','E','.')
	End;

	If @Player = @Show14 Then
	Begin
		If Not Optimizing And Not NoBrains Then
		Begin
			WriteLn('Warning: ','No optimization (-1) options needs the'^M^J+
							'         NoBrains -option (-L) with -QM. Set automatically.');
			NoBrains := True
		End;

		If Not AutoTicker Then
			WriteLn('Warning: ','It is',' recommended to use ','AutoTicker (-KA',') option with -','QM','. Not set.')
	End;

	If(@Player = @Show11)
	Or(@Player = @Show14)Then
		If UsedR Then
		Begin
			GammaValue := BasicGamma;
			If Debugging Then
				WriteLn('Warning: ','Gamma/Contrast value ',GammaValue,' set automatically'^M^J+
								'         to ',BasicGamma,' because',' -QM or -QF was specified.')
		End
		Else If Debugging Then
			WriteLn('Warning: ','Gamma/Contrast value ',GammaValue,' forced with -R'^M^J+
							'         switch and',' -QM or -QF was specified');

	If Optimizing And NoBrains
	And((@Player=@Show14)
		Or(@Player=@Show13)
		Or(@Player=@Show11)
		Or(@Player=@Show10)
		Or(@Player=@Show9)
		Or(@Player=@Show8)
		Or(@Player=@Show7)
		Or(@Player=@Show6)
		Or(@Player=@Show4))Then
		Error('*** In order to use NoBrains -option (-L), you'^M^J+
					'    must No Optimization (-1) option also when'^M^J+
					'    using either -QF,-QP,-QM,-QS,-J1,-J2,-J3,-A or -G.');

	If NoDetails > 4 Then
	Begin
		Error('*** Detail skipping level (-D) too big.'^M^J+
					'    Levels: -D    just skip details'^M^J+
					'            -DD   remove channel 3 + skip details'^M^J+
					'            -DDD  remove channels 2 and 3 + skip details'^M^J+
					'            -DDDD remove channels 0, 2 and 3 + skip details'^M^J)
	End;
	If NoDetails <> 0 Then
	Begin
		If NoBrains Then
			WriteLn('Warning: ','NoBrains -option (-L) together with'^M^J+
								 '         NoDetails -option (-D) makes HYPER brainlessness!');

		If Sustain > 0 Then
			Error('*** NoDetails -option (-D) is not compatible with'^M^J+
						'    Sustain   -option (-U).');

		If Not SkipVolDet Then
		Begin
			WriteLn('Warning: ','NoDetails (-D) options needs the'^M^J+
							'         SkipVolumeDetails -option (-KD). Set automatically.');
			SkipVolDet := True
		End;

		Write('Detaillessness level ',NoDetails);
		If Debugging Then
		Begin
			Write(':'^M^J'  - Slower reflexes on note changes'^M^J'  - ');
			If NoDetails > 1 Then Write('Remove')Else Write('Simplify');
			WriteLn(' channel 3');
			If NoDetails > 2 Then WriteLn('  - Remove channel 2');
			If NoDetails > 3 Then WriteLn('  - Remove channel 0')
		End
		Else WriteLn(' selected.')
	End;
	{$ENDIF}

	EnvInit(AIParam)
End;

End.
