jumps

;*      DRAWLINE.ASM
;*
;* S2 The Party '94 64kb intro
;* -- Line drawing routines
;*
;* Copyright 1995 Petteri Kangaslampi and Jarno Paananen
;*
;* This file is part of the MIDAS Sound System, and may only be
;* used, modified and distributed under the terms of the MIDAS
;* Sound System license, LICENSE.TXT. By continuing to use,
;* modify or distribute this file you indicate that you have
;* read the license and understand and accept it fully.
;*

include se.asm

data    segment word public
        extrn SVGAIncXRes:word
        extrn SVGAIncYRes:word

        extrn SCRXMax:word
        extrn SCRYMax:word
data    ends

code    segment byte public

        assume cs:code,ds:data

        public ClipLine

;       Outcodes used by line clipping routine:
;       ---------------------------------------
;         
; 010100011001
;         
; 
;     
; 010000001000
;     
; 
;         
; 011000101010
;         


;/***************************************************************************\
;*
;* Function:    ClipLine
;*
;* Description: Clips the endpoints of a line to the screen
;*
;* Input:       bx                      x-coordinate for point 1
;*              cx                      y-coordinate for point 1
;*              si                      x-coordinate for point 2
;*              di                      y-coordinate for point 2
;*
;* Returns:     ax = 1 if the line is completely out of the screen, 0 if not.
;*              line endpoint coordinates are updated
;*
;* Destroys:
;*
;\***************************************************************************/

ClipLine:
        push    bp

@@doclip:
        cmp     bx,si                   ; is the line vertical?
        je      @@vert
        cmp     cx,di
        je      @@horz                  ; or horizontal?

        xor ax, ax
        test cx,cx
        jns @@no1
        or al, 0001b
        jmp @@dox
@@no1:  cmp cx, SCRYMax
        jl @@dox
        or al, 0010b
@@dox:  test bx, bx
        jns @@no2
        or al, 0100b
        jmp @@no3
@@no2:  cmp bx, SCRXMax
        jl @@no3
        or al, 1000b
@@no3:  test di,di
        jns @@qo1
        or ah, 0001b
        jmp @@qox
@@qo1:  cmp di, SCRYMax
        jl @@qox
        or ah, 0010b
@@qox:  test si, si
        jns @@qo2
        or ah, 0100b
        jmp @@qo3
@@qo2:  cmp si, SCRXMax
        jl @@qo3
        or ah, 1000b
@@qo3:
        ; AL = OutCode for first point
        ; AH = OutCode for second point

        mov     dl,al                   ; don't draw if the line is fully
        and     dl,ah                   ; outside the screen (outcodes
        jnz     @@outside               ; have at least 1 same bit set

        mov     dl,al
        or      dl,ah                   ; if both outcodes are zero, the line
        jz      @@done                  ; is completely on screen

        test    al,al
        jnz     @@cp1
        mov     al,ah                   ; swap points if point 1 doesn't need
        xchg    bx,si                   ; clipping
        xchg    cx,di

@@cp1:  mov     bp,ax                   ; bp = outcode for point 1

        test    al,0001b
        jz      @@no11

        mov     ax,si                   ; outcode AND 1 != 0 --> point is
        sub     ax,bx                   ; above the window
        neg     cx
        imul    cx
        add     cx,di                   ; x1 = x1 + (x2-x1) * (0-y1) / (y2-y1)
        idiv    cx
        add     bx,ax

        xor     cx,cx                   ; y1 = 0

        jmp     @@no12


@@no11: test    bp,0010b
        jz      @@no12

        mov     ax,si                   ; outcode AND 2 != 0 --> point is
        sub     ax,bx                   ; below the window
        mov     dx,SCRYMax
        dec dx
        sub     dx,cx
        imul    dx
        neg     cx
        add     cx,di                   ; x1 = x1 + (x2-x1) * (yMax-y1)
        idiv    cx                      ;       / (y2-y1)
        add     bx,ax

        mov     cx,SCRYMax              ; y1 = yMax
        dec cx


@@no12: cmp     bx,si
        je      @@horz

        test    bp,0100b
        jz      @@no13

        mov     ax,di                   ; outcode AND 4 != 0 --> point is
        sub     ax,cx                   ; to left of the window
        neg     bx
        imul    bx
        add     bx,si
        idiv    bx                      ; y1 = y1 + (y2-y1) * (0-x1) / (x2-x1)
        add     cx,ax

        xor     bx,bx                   ; x1 = 0

        jmp     @@doclip


@@no13: test    bp,1000b
        jz      @@doclip

        mov     ax,di                   ; outcode AND 8 != 0 --> point is
        sub     ax,cx                   ; to right of the window
        mov     dx,SCRXMax
        dec dx

        sub     dx,bx
        imul    dx
        neg     bx
        add     bx,si
        idiv    bx                      ; y1 = y1 + (y2-y1) * (xMax-x1)
        add     cx,ax                   ;       / (x2-x1)

        mov     bx,SCRXMax              ; x1 = xMax
        dec bx

        jmp     @@doclip

@@outside:
        mov     ax,1
        jmp     @@exit


@@vert: ; vertical line
        cmp     cx,di
        jle     @@v1                    ; swap endpoints if y2<y1
        xchg    cx,di
        xchg    bx,si

@@v1:   test    di,di
        js      @@outside               ; is the line completely outside the
        cmp     cx,SCRYMax              ; screen?
        jge     @@outside
        cmp     bx,SCRXMax              ; in X direction?
        jae     @@outside

        cmp     di,SCRYMax
        jl      @@v2
        mov     di,SCRYMax
        dec di

@@v2:   test    cx,cx
        jns     @@done
        xor     cx,cx
        jmp     @@done


@@horz: ; horizontal line
        cmp     bx,si
        jle     @@h1                    ; swap endpoints if x2<x1
        xchg    bx,si
        xchg    cx,di

@@h1:   test    si,si
        js      @@outside               ; is the line completely outside the
        cmp     bx,SCRXMax              ; screen?
        jge     @@outside
        cmp     cx,SCRYMax              ; in Y direction?
        jae     @@outside

        cmp     si,SCRXMax
        jl      @@h2
        mov     si,SCRXMax
        dec si

@@h2:   test    bx,bx
        jns     @@done
        xor     bx,bx


@@done: xor     ax,ax

@@exit: pop     bp

        retf

code    ends

END
