unit Strings;

{$O+,D+,S-,R-,Q-,I-}
interface

{ StrLen returns the number of characters in Str, not counting  }
{ the null terminator.}
Function StrLen(Str: PChar): Word;

{ StrEnd returns a pointer to the null character that           }
{ terminates Str.}
Function StrEnd(Str: PChar): PChar;

{ StrMove copies exactly Count characters from Source to Dest   }
{ and returns Dest. Source and Dest may overlap.}
Function StrMove(Dest, Source: PChar; Count: Word): PChar;

{ StrCpy copies Source to Dest and returns Dest.}
Function StrCpy(Dest, Source: PChar): PChar;

{ StrECpy copies Source to Dest and returns StrEnd(Dest).}
Function StrECpy(Dest, Source: PChar): PChar;

{ strnCpy copies at most MaxLen characters from Source to Dest }
{ and returns Dest.}
Function strnCpy(Dest, Source: PChar; MaxLen: Word): PChar;

{ StrPCpy copies the Pascal style string Source into Dest and  }
{ returns Dest.}
Function StrPCpy(Dest: PChar; Source: String): PChar;

{ StrCat appends a Cpy of Source to the End of Dest and        }
{ returns Dest.}
Function StrCat(Dest, Source: PChar): PChar;

{ strnCat appends at most MaxLen - StrLen(Dest) characters from }
{ Source to the End of Dest, and returns Dest.}
Function strnCat(Dest, Source: PChar; MaxLen: Word): PChar;

{ StrCmp compares Str1 to Str2. The return value is less than  }
{ 0 if Str1 < Str2, 0 if Str1 = Str2, or greater than 0 if      }
{ Str1 > Str2.}
Function StrCmp(Str1, Str2: PChar): Integer;

{ StrICmp compares Str1 to Str2, without case sensitivity. The }
{ return value is the same as StrCmp.}
Function StrICmp(Str1, Str2: PChar): Integer;

{ StrnCmp compares Str1 to Str2, for a maximum length of       }
{ MaxLen characters. The return value is the same as StrCmp.}
Function StrnCmp(Str1, Str2: PChar; MaxLen: Word): Integer;

{ StrnICmp compares Str1 to Str2, for a maximum length of      }
{ MaxLen characters, without case sensitivity. The return value }
{ is the same as StrCmp.}
Function StrnICmp(Str1, Str2: PChar; MaxLen: Word): Integer;

{ StrScan returns a pointer to the first occurrence of Chr in   }
{ Str. If Chr does not occur in Str, StrScan returns NIL. The   }
{ null terminator is considered to be part of the string.}
Function StrScan(Str: PChar; Chr: Char): PChar;

{ StrRScan returns a pointer to the last occurrence of Chr in   }
{ Str. If Chr does not occur in Str, StrRScan returns NIL. The  }
{ null terminator is considered to be part of the string.}
Function StrRScan(Str: PChar; Chr: Char): PChar;

{ StrPos returns a pointer to the first occurrence of Str2 in   }
{ Str1. If Str2 does not occur in Str1, StrPos returns NIL.}
Function StrPos(Str1, Str2: PChar): PChar;

{ StrUpper converts Str to upper case and returns Str.}
Function StrUpper(Str: PChar): PChar;

{ StrLower converts Str to lower case and returns Str.}
Function StrLower(Str: PChar): PChar;

{ StrPas converts Str to a Pascal style string.}
Function StrPas(Str: PChar): String;

{ StrNew allocates a Cpy of Str on the heap. If Str is NIL or  }
{ points to an empty string, StrNew returns NIL and doesn't     }
{ allocate any heap space. Otherwise, StrNew makes a duplicate  }
{ of Str, obtaining space with a call to the GetMem standard    }
{ Procedure, and returns a pointer to the duplicated string.    }
{ The allocated space is StrLen(Str) + 1 bytes long.}
Function StrNew(Str: PChar): PChar;

{ StrDispose disposes a string that was previously allocated    }
{ with StrNew. If Str is NIL, StrDispose does nothing.}
Procedure StrDispose(Str: PChar);

{Bisqwit procedures}

{ StrUpper converts Str to upper case and returns Str.}
Function StrnUpper(Str: PChar; Max: Word): PChar;

{ StrnPos returns a pointer to the first occurrence of Str2 in }
{ Str1. If Str2 does not occur in Str1, StrnPos returns NIL. }
Function StrNPos(Str1, Str2: PChar; Max: LongInt): PChar;

{ StrnPos returns a pointer to the first occurrence of Str2 in }
{ Str1. If Str2 does not occur in Str1, StrnPos returns NIL. }
Function StrNIPos(Str1, Str2: PChar; Max: LongInt): PChar;

{Returns count of What in Str}
Function StrCount(Str: PChar; What: Char): Word;

{Returns count of What in Str (max len = MaxLen)}
Function StrnCount(Str: PChar; What: Char; MaxLen: Word): Word;

{Deletes a substring from a string.}
Function StrDel(Str: PChar; Index, Count: Word): PChar;

{ StrCCat appends Dest by Chr and returns Dest.}
Function StrCCat(Dest: PChar; Chr: Char): PChar;

{ChrPChar converts Ch to PChar}
Function ChrPChar(Chr: Char): PChar;

{ StrScan returns a pointer to the first occurrence of Chr in Str. }
{ If Chr does not occur in first Max characters of Str, StrScan returns }
{ NIL. The null terminator is considered to be part of the string. }
Function StrnScan(Str: PChar; Chr: Char; Max:Word): PChar;

Function StrSpFix(s: PChar): PChar;

Function MemCmp(Str1, Str2: PChar; MaxLen: Word): Integer;
Function MemEqual(Str1, Str2: PChar; MaxLen: Word): Word;
{ MemEqual returns equal byte count }

implementation

{$IFDEF USEDOSMEM}
Uses DosMem;
{$ENDIF}

{$W-}
Function StrLen(Str: PChar): Word; assembler;
Asm
        CLD
        LES     DI,Str
        MOV     CX,0FFFFH
        XOR     AL,AL
        REPNE   SCASB
        MOV     AX,0FFFEH
        SUB     AX,CX
End;

Function StrEnd(Str: PChar): PChar; assembler;
Asm
				CLD
        LES     DI,Str
        MOV     CX,0FFFFH
        XOR     AL,AL
        REPNE   SCASB
        MOV     AX,DI
        MOV     DX,ES
        DEC     AX
End;

Function StrMove(Dest, Source: PChar; Count: Word): PChar; assembler;
Asm
        PUSH    DS
         CLD
         LDS     SI,Source
         LES     DI,Dest
         MOV     AX,DI
         MOV     DX,ES
         MOV     CX,Count
         CMP     SI,DI
         JAE     @@1
         STD
         ADD     SI,CX
         ADD     DI,CX
         DEC     SI
         DEC     DI
@@1:     REP     MOVSB
         CLD
        POP     DS
End;

Function StrCpy(Dest, Source: PChar): PChar; assembler;
Asm
        PUSH    DS
         CLD
         LES     DI,Source
         MOV     CX,0FFFFH
         XOR     AL,AL
         REPNE   SCASB
         NOT     CX
         LDS     SI,Source
         LES     DI,Dest
         MOV     AX,DI
         MOV     DX,ES
         REP     MOVSB
        POP     DS
End;

Function StrECpy(Dest, Source: PChar): PChar; assembler;
Asm
        PUSH    DS
        CLD
        LES     DI,Source
        MOV     CX,0FFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     CX
        LDS     SI,Source
        LES     DI,Dest
        REP     MOVSB
        MOV     AX,DI
        MOV     DX,ES
        DEC     AX
        POP     DS
End;

Function StrnCpy(Dest, Source: PChar; MaxLen: Word): PChar; assembler;
Asm
				PUSH    DS
				 CLD
				 LES     DI,Source
				 MOV     CX,MaxLen
				 MOV     BX,CX
				 XOR     AL,AL
				 REPNE   SCASB
				 SUB     BX,CX
				 MOV     CX,BX
				 LDS     SI,Source
				 LES     DI,Dest
				 MOV     BX,DI
				 MOV     DX,ES
				 REP     MOVSB
				 STOSB
				 XCHG    AX,BX
        POP     DS
End;

Function StrPCpy(Dest: PChar; Source: String): PChar; assembler;
Asm
        PUSH    DS
        CLD
        LDS     SI,Source
        LES     DI,Dest
        MOV     BX,DI
        MOV     DX,ES
        LODSB
        XOR     AH,AH
        XCHG    AX,CX
        REP     MOVSB
        XOR     AL,AL
        STOSB
        XCHG    AX,BX
        POP     DS
End;

Function StrCat(Dest, Source: PChar): PChar; assembler;
Asm
        PUSH    Dest.Word[2]
        PUSH    Dest.Word[0]
        PUSH    CS
        CALL    NEAR PTR StrEnd
        PUSH    DX
        PUSH    AX
        PUSH    Source.Word[2]
        PUSH    Source.Word[0]
        PUSH    CS
        CALL    NEAR PTR StrCpy
        MOV     AX,Dest.Word[0]
        MOV     DX,Dest.Word[2]
End;

Function StrCCat(Dest: PChar; Chr: Char): PChar; assembler; {Bisqwit}
Asm
        PUSH    Dest.Word[2]
        PUSH    Dest.Word[0]
        PUSH    CS
        CALL    NEAR PTR StrEnd
        mov es, dx
				mov di, ax
				xor ah, ah
				mov al, Chr
				mov es:[di], ax

        MOV AX,Dest.Word[0]
        MOV DX,Dest.Word[2]
End;

Function StrnCat(Dest, Source: PChar; MaxLen: Word): PChar; assembler;
Asm
        PUSH    Dest.Word[2]
        PUSH    Dest.Word[0]
        PUSH    CS
        CALL    NEAR PTR StrEnd
        MOV     CX,Dest.Word[0]
        ADD     CX,MaxLen
        SUB     CX,AX
        JBE     @@1
        PUSH    DX
        PUSH    AX
        PUSH    Source.Word[2]
        PUSH    Source.Word[0]
        PUSH    CX
        PUSH    CS
        CALL    NEAR PTR StrnCpy
@@1:    MOV     AX,Dest.Word[0]
        MOV     DX,Dest.Word[2]
End;

Function StrCmp(Str1, Str2: PChar): Integer; assembler;
Asm
        PUSH    DS
        CLD
        LES     DI,Str2
        MOV     SI,DI
        MOV     CX,0FFFFH
        XOR     AX,AX
        CWD
        REPNE   SCASB
        NOT     CX
        MOV     DI,SI
        LDS     SI,Str1
        REPE    CMPSB
        MOV     AL,DS:[SI-1]
        MOV     DL,ES:[DI-1]
        SUB     AX,DX
        POP     DS
End;

Function StrICmp(Str1, Str2: PChar): Integer; assembler;
Asm
        PUSH    DS
        CLD
        LES     DI,Str2
        MOV     SI,DI
        MOV     CX,0FFFFH
        XOR     AX,AX
        CWD
        REPNE   SCASB
        NOT     CX
        MOV     DI,SI
        LDS     SI,Str1
@@1:    REPE    CMPSB
        JE      @@4
        MOV     AL,DS:[SI-1]
        CMP     AL,'a'
        JB      @@2
        CMP     AL,'z'
        JA      @@2
        SUB     AL,20H
@@2:    MOV     DL,ES:[DI-1]
        CMP     DL,'a'
        JB      @@3
        CMP     DL,'z'
        JA      @@3
        SUB     DL,20H
@@3:    SUB     AX,DX
        JE      @@1
@@4:    POP     DS
End;

Function StrnCmp(Str1, Str2: PChar; MaxLen: Word): Integer; assembler;
Asm
        PUSH    DS
        CLD
        LES     DI,Str2
        MOV     SI,DI
        MOV     AX,MaxLen
        MOV     CX,AX
        JCXZ    @@1
        XCHG    AX,BX
        XOR     AX,AX
        CWD
        REPNE   SCASB
        SUB     BX,CX
        MOV     CX,BX
        MOV     DI,SI
        LDS     SI,Str1
        REPE    CMPSB
        MOV     AL,DS:[SI-1]
        MOV     DL,ES:[DI-1]
        SUB     AX,DX
@@1:    POP     DS
End;

Function StrnICmp(Str1, Str2: PChar; MaxLen: Word): Integer; assembler;
Asm
        PUSH    DS
        CLD
        LES     DI,Str2
        MOV     SI,DI
        MOV     AX,MaxLen
        MOV     CX,AX
        JCXZ    @@4
        XCHG    AX,BX
        XOR     AX,AX
        CWD
        REPNE   SCASB
        SUB     BX,CX
        MOV     CX,BX
        MOV     DI,SI
        LDS     SI,Str1
@@1:    REPE    CMPSB
        JE      @@4
        MOV     AL,DS:[SI-1]
        CMP     AL,'a'
        JB      @@2
        CMP     AL,'z'
        JA      @@2
        SUB     AL,20H
@@2:    MOV     DL,ES:[DI-1]
        CMP     DL,'a'
        JB      @@3
        CMP     DL,'z'
        JA      @@3
        SUB     DL,20H
@@3:    SUB     AX,DX
        JE      @@1
@@4:    POP     DS
End;

Function StrScan(Str: PChar; Chr: Char): PChar; assembler;
Asm
        CLD
        LES     DI,Str
        MOV     SI,DI
        MOV     CX,0FFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     CX
        MOV     DI,SI
        MOV     AL,Chr
        REPNE   SCASB
        MOV     AX,0
        CWD
        JNE     @@1
        MOV     AX,DI
        MOV     DX,ES
        DEC     AX
@@1:
End;

Function StrRScan(Str: PChar; Chr: Char): PChar; assembler;
Asm
        CLD
        LES     DI,Str
        MOV     CX,0FFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     CX
        STD
        DEC     DI
        MOV     AL,Chr
        REPNE   SCASB
        MOV     AX,0
        CWD
        JNE     @@1
        MOV     AX,DI
        MOV     DX,ES
        INC     AX
@@1:    CLD
End;

Function StrPos(Str1, Str2: PChar): PChar;
Begin
  StrPos := StrNPos(Str1, Str2, strlen(str1))
End;

Function StrUpper(Str: PChar): PChar; assembler;
Asm
        PUSH    DS
         lds dx, Str
         mov ax, 6522h
         int 21h
;         CLD
;         LDS     SI,Str
;         MOV     BX,SI
;         MOV     DX,DS
;@@1:     LODSB
;         OR      AL,AL
;         JE      @@2
;         CMP     AL,'a'
;         JB      @@1
;         CMP     AL,'z'
;         JA      @@1
;         SUB     AL,20H
;         MOV     [SI-1],AL
;         JMP     @@1
;@@2:     XCHG    AX,BX
        POP     DS
End;

Function StrLower(Str: PChar): PChar; assembler;
Asm
        PUSH    DS
         CLD
         LDS     SI,Str
         MOV     BX,SI
         MOV     DX,DS
@@1:     LODSB
         OR      AL,AL
         JE      @@2
         CMP     AL,'A'
         JB      @@1
         CMP     AL,'Z'
         JA      @@1
         ADD     AL,20H
         MOV     [SI-1],AL
         JMP     @@1
@@2:     XCHG    AX,BX
        POP     DS
End;

Function StrPas(Str: PChar): String; assembler;
Asm
        PUSH    DS
         CLD
         LES     DI,Str
         MOV     CX,0FFFFH
         XOR     AL,AL
         REPNE   SCASB
         NOT     CX
         DEC     CX
         LDS     SI,Str
         LES     DI,@Result
         MOV     AL,CL
         STOSB
         REP     MOVSB
        POP     DS
End;

{$W+}
Function StrNew(Str: PChar): PChar;
Var
	L: Word;
	P: PChar;
Begin
	StrNew := Nil;
	If(Str <> Nil)And(Str^ <> #0)Then
	Begin
		L := StrLen(Str) + 1;
		{$IFDEF USEDOSMEM}
		AllocMem('', P, (L+15)Shr 4);
		{$ELSE}
		GetMem(P, L);
		{$ENDIF}
		If P <> Nil Then StrNew := StrMove(P, Str, L)
	End
End;

Procedure StrDispose(Str: PChar);
Begin
	If Str <> Nil Then
	{$IFDEF USEDOSMEM}
		UnAllocMem(Str)
	{$ELSE}
		FreeMem(Str, StrLen(Str) + 1)
	{$ENDIF}
End;

{Bisqwit procedures}

Function StrNPos(Str1, Str2: PChar; Max: LongInt): PChar;
Var
  Taulu: Array[Char]Of Byte;
  p: PChar;
  Ch: Char;
  a, b: Byte;
Begin
  If Str2^ = #0 Then
  Begin
    StrNPos := Str1;
    Exit
  End;
  b := strlen(Str2);
  FillChar(Taulu, SizeOf(Taulu), b);
  Dec(b);
  For Ch := #1 To #255 Do
  Begin
    p := strrscan(Str2, Ch);
		If p<>Nil Then
			Taulu[Ch] := Min(1, b - (p-Str2));
	End;

	p := Str1;
	Repeat
		a := b;
		While p[a] = Str2[a]Do
		Begin
			If a=0 Then
			Begin
				StrNPos := p;
				Exit
			End;
			Dec(a)
		End;
		If Max < Taulu[p[a]]Then Break;

		Dec(Max, Taulu[p[a]]);
		Inc(p, System.Max(1, Taulu[p[a]]))
	Until False;

	StrNPos := Nil
End;

Function StrNIPos(Str1, Str2: PChar; Max: LongInt): PChar;
Var
  Taulu: Array[Char]Of Byte;
  p: PChar;
  Ch: Char;
  a, b: Byte;
Begin
  If Str2^ = #0 Then
  Begin
    StrNIPos := Str1;
    Exit
  End;
  b := strlen(Str2);
  If(Str1=Nil)Or(Max < b)Then
  Begin
    StrNIPos := Nil;
    Exit
  End;
  FillChar(Taulu, SizeOf(Taulu), b);
  Dec(b);
  For Ch := #1 To #255 Do
  Begin
    p := strrscan(Str2, UpCase(Ch));
    If p<>Nil Then Taulu[Ch] := Min(1, b - (p-Str2))
  End;

  p := Str1;
  Repeat
    a := b;
    While UpCase(p[a]) = UpCase(Str2[a])Do
    Begin
      If a=0 Then
      Begin
        StrNIPos := p;
        Exit
      End;
      Dec(a)
    End;
    If Max < Taulu[p[a]]Then Break;

    Dec(Max, Taulu[p[a]]);
    Inc(p, System.Max(1,Taulu[p[a]]))
  Until False;

  StrNIPos := Nil
End;

Function StrCount(Str: PChar; What: Char): Word; Assembler;
Asm
				cld
				les si, Str
				xor cx, cx
@@1:    seges lodsb
				or al, al
				jz @@3
				cmp al, What
				jnz @@1
				inc cx
				jmp @@1
@@3:    mov ax, cx
End;

Function StrNCount(Str: PChar; What: Char; MaxLen: Word): Word; Assembler;
Asm
				cld
				les si, Str
				mov bx, MaxLen
				xor cx, cx
@@1:    or bx, bx
				jz @@3
				dec bx
				seges lodsb
				cmp al, What
				jnz @@1
				inc cx
				jmp @@1
@@3:    mov ax, cx
End;

Function StrDel(Str: PChar; Index,Count: Word): PChar; Assembler;
Asm
				CLD
				LES     DI,Str
				MOV     CX,0FFFFH
				XOR     AL,AL
				REPNE   SCASB
				NOT     CX
				DEC     CX
				mov ax, Index
				cmp cx, ax
				jb @@9
				mov di, Str.word[0]
				add di, ax

				mov bx, Count
@@1:    mov al, byte ptr es:[di]
				or al, al
				jz @@9
				mov al, byte ptr es:[di+bx]
				stosb
				or al, al
				jnz @@1

@@9:    mov ax, Str.word[0]
				mov dx, Str.word[2]
End;

Function ChrPChar(Chr: Char): PChar; Assembler;
Asm
  mov bx, offset @@1
	{$IFDEF DPMI}
    mov ax, cs
    add ax, SelectorInc
    mov es, ax
    mov al, Chr
    mov byte ptr es:[bx], al
  {$ELSE}
    mov al, Chr
    mov byte ptr cs:[bx], al
  {$ENDIF}
  jmp @@2
@@1:dw 0
@@2:
  mov dx, cs
  mov ax, bx
End;

Function StrnUpper(Str: PChar; Max: Word): PChar; assembler;
Asm
        PUSH    DS
         lds dx, Str
         mov cx, Max
         mov ax, 6521h
         int 21h
;         CLD
;         LDS     SI,Str
;         MOV     BX,SI
;         MOV     DX,DS
;@@1:     LODSB
;         OR      AL,AL
;         JE      @@2
;         CMP     AL,'a'
;         JB      @@1
;         CMP     AL,'z'
;         JA      @@1
;         SUB     AL,20H
;         MOV     [SI-1],AL
;         JMP     @@1
;@@2:     XCHG    AX,BX
        POP     DS
End;

Function StrnScan(Str: PChar; Chr: Char; Max:Word): PChar; assembler;
Asm
        CLD
				LES     DI,Str
        mov cx, Max
        MOV     AL,Chr
        REPNE   SCASB
        MOV     AX,0
        CWD
        JNE     @@1
        MOV     AX,DI
        MOV     DX,ES
        DEC     AX
@@1:
End;

Function StrSpFix(s: PChar): PChar; Assembler;
Asm
	les di, s
	push ds
	 lds si, s
@1:
	 mov al, ds:[si]
	 cmp al, ' '
	 jne @2
	 inc si
	 jmp @1
@2:
	 cmp al, 0
	 jz @3a
{	 inc si}
{	 inc di}
	 mov al, ds:[si]
	 mov es:[di], al
	 inc si
	 inc di
	 jmp @2
@3a:
	 dec di
@3:
	 cmp di, s.word[0]
	 jbe @5
	 dec di
	 cmp byte ptr es:[di], ' '
	 je @3
	 inc di
@5:
	 mov es:[di], al
	pop ds
	mov ax, s.word[0]
	mov dx, es
End;

Function MemCmp(Str1, Str2: PChar; MaxLen: Word): Integer; assembler;
Asm
				PUSH    DS
				 CLD
				 LES     DI,Str2
				 MOV     AX,MaxLen
				 MOV     CX,AX
				 JCXZ    @@1
				 LDS     SI,Str1
				 REPE    CMPSB
				 MOV     AL,DS:[SI-1]
				 MOV     DL,ES:[DI-1]
				 SUB     AX,DX
				 xor ah, ah
@@1:    POP     DS
End;

Function MemEqual(Str1, Str2: PChar; MaxLen: Word): Word; assembler;
Asm
				PUSH    DS
				 CLD
				 LES     DI,Str2

				 xor ax, ax
				 mov cx, MaxLen

				 lds si, Str1
				 jcxz @@1
@@2:		 cmpsb
				 jnz @@1
				 inc ax
				 loop @@2
@@1:    POP     DS
End;

End.
